#ifndef FWI_ABSTRACTINTERFACEDGTSVA
#define FWI_ABSTRACTINTERFACEDGTSVA

#include "DGViscoAcousticProblem.hpp"

class FWIMaterialVAcousticDGTS  {
    int numL;
    int paramComps;
  std::string parametrization;
protected:
    bool initial_value;
public:
    explicit FWIMaterialVAcousticDGTS(int _numL = 1)
            : numL(_numL), initial_value(false), paramComps(0) {}

    virtual void setInitialValueFlag(bool ivf) { initial_value = ivf; }

    [[nodiscard]] int GetNumL() const { return numL; }

    [[nodiscard]] int GetParamComps() const {
        if (parametrization.empty()) {
            Exit("make sure you set the parametrization in the problem!!")
        }
        return paramComps;
    }

    [[nodiscard]] virtual double Kappa_i(const Point &, int) const {
        Exit("not specified");
        return 0.0;
    }

    [[nodiscard]] virtual double Rho(const Point &) const {
        Exit("not specified");
        return 0.0;
    }

    [[nodiscard]] virtual double Kappa(const Cell &C, const Point &x) const {
        double kappa = 0.0;
        for (int i = 0; i <= numL; ++i) {
            kappa += Kappa_i(C, x, i);
        }
        return kappa;
    }

    [[nodiscard]] virtual double Kappa_i(const Cell &, const Point &x, int i) const {
        return Kappa_i(x,
                       i);
    }

  [[nodiscard]] virtual double Tau_i(const Cell &C, const Point &x, int i) const {
        THROW("Tau_i not implemented");
        return 0.0;
    }

    [[nodiscard]] virtual double invTau_i(const Cell &C, const Point &x, int i) const {
        return 1.0 / Tau_i(C, x, i);
    }

    [[nodiscard]] virtual double Rho(const Cell &, const Point &x) const { return Rho(x); }

    [[nodiscard]] virtual double alpha(const Cell &,
                         const Point &) const {
      THROW("alpha not implemented");
    }
    [[nodiscard]] virtual double alpha1() const {
      THROW("alpha1 not implemented");
    }
    [[nodiscard]] virtual double alpha2() const {
      THROW("alpha2 not implemented");
    }

    [[nodiscard]] virtual double V_p(const Cell &, const Point &) const {
      THROW("V_p not implemented");
    }

    [[nodiscard]] virtual double Tau_p(const Cell &,
                         const Point &) const {
      THROW("Tau_p not implemented");
    }

    void setParametrization(const std::string &s) {
      parametrization = s;
      if (s == "native") {
        paramComps = 2 + 2 * numL;
      } else if (s == "taup" || s == "taup_positive") {
        paramComps = (numL==0) ? 2 : 3+numL;
      }
    }
    [[nodiscard]] virtual const std::string &getParametrization() const { return parametrization; }

    ~FWIMaterialVAcousticDGTS() = default;
};

class FWIBndConditionsVAcoustic {
    int dim;
    int numL;
public:
    FWIBndConditionsVAcoustic(int _dim, int _numL)
        : dim(_dim), numL(_numL) {
    }

    int GetStateComps() const { return dim + 1 + numL; }

    int GetDim() const { return dim; }

    double Dirichlet_i(double t, const Point &z, int i) const {
        return ut(t,z, dim + i);
    }

    double Neumann(double t, const Cell &C, const Point &z, const VectorField &N) const{
      VectorField V = zero;
      for (int i = 0; i < dim; ++i) {
        V[i] = ut(t, C, z, i);
      }
      return V * N;
    }
    virtual double ut(double, const Point &, int) const {
      THROW("ut not implemented");
    }

    virtual double ut(double t, const Cell &, const Point &z, int i) const {
        return ut(t,z,i);
    }

    virtual bool exact() const {
      THROW("exact not implemented");
    }
};

class FWIRhsVAcoustic  {
protected:
  int shot_count = 1;
public:
  [[nodiscard]] virtual int GetShotCount() const { return shot_count; }
  virtual void SetShot(int _shot) {Exit("CIAO"); }

  virtual int getNrPointSrc() const {
      THROW("getNrPointSrc not implemented");
    }

    virtual Point getPointSrcLoc(int /*nr*/) const {
      THROW("getPointSrcLoc not implemented");
    };

    virtual bool RHS(double t) const {
      THROW("RHS not implemented");
    }

    virtual VectorField ForceV(double, const Cell &, const Point &, int = -1) const {
      THROW("ForceV not implemented");
    }

    virtual double ForceP_i(double, const Cell &, const Point &, int, int = -1) const {
      THROW("ForceP_i not implemented");
    }

    virtual void Force(VectorField &V_rhs,
                       RVector &Force,
                       double t,
                       const Cell &C,
                       const Point &z,
                       int q = -1) const {
        V_rhs = ForceV(t, C, z, q);
        for (int n = 0; n < Force.size(); ++n) {
            Force[n] = ForceP_i(t, C, z, n, q);
        }
    }
    virtual ~FWIRhsVAcoustic() = default;
};

class WaveProblemDGTSVA
        : public FWIBndConditionsVAcoustic, public FWIRhsVAcoustic, public FWIMaterialVAcousticDGTS {
protected:
public:
    WaveProblemDGTSVA(int _numL, int _dim) : FWIBndConditionsVAcoustic(_dim, _numL),
                                             FWIRhsVAcoustic(),
                                             FWIMaterialVAcousticDGTS(_numL) {}

    int GetStateComponents() {
        return GetDim() + 1 + GetNumL();
    }

    using FWIRhsVAcoustic::getNrPointSrc;

    virtual const char *Name() const {
      THROW("Name not implemented");
    }
};


#endif
