#ifndef FWIFORWARDMODEL
#define FWIFORWARDMODEL

#include <memory>
#include <unordered_map>
#include <utility>
#include <vector>

#include "Config.hpp"
#include "DGVAcousticAssembleBwd.hpp"
#include "DGVAcousticPlotter.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "FWIProblemWrapper.hpp"
#include "FWITaper.hpp"
#include "FWITimeSolver.hpp"
#include "Helper/ABCHandler.hpp"
#include "Helper/FWIDataContainerDomain.hpp"
#include "Helper/FWIDataContainerRange.hpp"
#include "Helper/FWIHelpers.hpp"
#include "Helper/ReconstructionContainer.hpp"
#include "Helper/VectorFillFunctions.hpp"
#include "Observation/MeasureMask.hpp"
#include "Problems/FWIAcousticInversionProb.hpp"
#include "VAcousticMatFromVec.hpp"
#include "lib4_fem/elements/Element.hpp"

typedef RVector FWIState;
class ForwardModelPlotter {
  DGVAcousticPlotter Plotter;
  int plotFactor = 1;
  int plotEnding{};
  bool allowPlotting_M = true;
  bool allowPlotting_WF = false;
  int numL{};
  const ReconstructionContainer& recon;
  const std::string &param;

public:
  ForwardModelPlotter(
      const ReconstructionContainer &recon_,
      const std::unordered_map<Point, DGAcousticElement *> *EleCache,
      int nSteps, int numL_, const std::string &param_)
      : Plotter(numL_, EleCache), recon(recon_), param(param_) {
    numL = numL_;
    Config::Get("allowPlotting_M", allowPlotting_M);
    Config::Get("allowPlotting_WF", allowPlotting_WF);
    Config::Get("plotFactor", plotFactor);
    plotEnding = static_cast<int>(round(nSteps / plotFactor)) + 1;
  }
    void Plot(const FWIDataContainerDomain &m, const std::string &name, int nr) {
      const Vector &M = m.getMat();
      if (!allowPlotting_M)
        return;
      std::string filename;
      filename = nr_str(name.c_str(), {nr});
      plotMatVector(M, filename);
      if (m.isWaveField())
        plot(m.getVectors(), (filename + "_WF"));
    }
    void plotComponent(const Vector &M, const std::string &component, int index,
                       const std::string &filename) {
      auto &Plot = mpp::plot(component + "_" + filename);
      Plot.AddData(component, M, index);
      Plot.PlotFile();
    }
    void plotMatVector(const Vector &M,
                       const std::string &filename) {
      LagrangeDiscretization disc(M.GetDisc().GetMeshes(), 0);
      if (recon.Rho || Plotter.opt.plot_rho) { plotComponent(M, "Rho", 0, filename);
      }
      if (param == "native") {
        if (recon.Kappa0) {
          for (int i = 0; i <= numL; ++i) {
            std::string kappaName = "Kappa" + std::to_string(i);
            plotComponent(M, kappaName, i + 1, filename);
          }
        }
      } else if (param == "taup") {
        if (recon.Vp) {
          plotComponent(M, "Vp", 1, filename);
        }
        if (recon.Taup && numL > 0) {
          plotComponent(M, "Taup", 2, filename);
        }
      } else if (param == "taup_positive") {
        Vector M2(M);
        int taupPlotLevel = 0;
        Config::Get("taupPlotLevel", taupPlotLevel);
        for (cell c = M2.cells(); c != M2.cells_end(); ++c) {
          int _comps = 2 + (numL > 0);
          for (int i = 0; i < _comps; ++i) {
            M2(c(), i) = exp(M(c(), i));
          }
        }
        if (recon.Vp) {
          if (taupPlotLevel >= 1) {
            plotComponent(M, "Vp", 1, filename);
          }
          plotComponent(M2, "VpExp", 1, filename);
        }
        if (recon.Taup && numL > 0) {
          if (taupPlotLevel >= 1) {
            plotComponent(M, "Taup", 2, filename);
          }
          plotComponent(M2, "TaupExp", 2, filename);
        }
        if (recon.Rho)
          plotComponent(M2, "RhoExp", 0, filename);
      } else {
        Exit("Unknown parameterization");
      }
    }
    void plot(const Vector &M, const std::string &name,int nr) {
      if (!allowPlotting_M)
        return;
      std::string filename;
      filename = nr_str(name.c_str(), {nr});
      plotMatVector(M, filename);
    };

    void plot(const Vectors &U, const std::string &name) {
      if (!allowPlotting_WF) return;
      for (int i = 0; i < plotEnding; ++i) {
        Plotter.plotWaveVector(U[i * plotFactor], i, name.c_str());
      }
    }
};
class ForwardModel {
protected:
  int deg{};
  std::vector<std::unique_ptr<Vectors>> uM;
  ReconstructionContainer recon;
  FWIDiscretizationStack FWIDiscStack;
  std::shared_ptr<FWIDiscretizationStack> FWIDiscStackPTR;
    bool useGardener = false;
  std::shared_ptr<EleCacheContainer> eleCacheContainer;

  std::unique_ptr<MeasureMask> Mask;
  std::shared_ptr<ABCHandler> ABCH;
  void initializeMeasureMask() {
    int MeasureComp = 0;
    Config::Get("MeasureComp", MeasureComp);
    Mask = GetMeasureMask(problem_fwd->GetDim(), problem_fwd->GetNumL(),
                          MeasureComp);
  }
public:
  static constexpr double Eps = 1e-10;

  std::unique_ptr<FWIObservationOperatorDGTS> Obs;
  std::unique_ptr<FWITaper> Taper;
  bool precon{false};
  double PreconRegu{1e-8};
  std::shared_ptr<FWIVAcousticInversionProb> problem_fwd;
  TimeConf TConf;
  ForwardModelPlotter Plotter;

  ForwardModel(std::shared_ptr<FWIVAcousticInversionProb> problem_fwd_,
               std::shared_ptr<FWIDiscretizationStack> FWIDiscStackPTR_,
               std::shared_ptr<EleCacheContainer> eleCacheContainer_,
               std::unique_ptr<FWIObservationOperatorDGTS> Obs_,
               std::unique_ptr<FWITaper> Taper_
               ) :
  recon(*problem_fwd_),
  Plotter(recon,eleCacheContainer->Get().get(),
          TConf.timeSteps,problem_fwd_->GetNumL(),
                problem_fwd_->getParametrization()),
  Obs(std::move(Obs_)),
  eleCacheContainer(std::move(eleCacheContainer_)),
        FWIDiscStackPTR(std::move(FWIDiscStackPTR_)),
        Taper(std::move(Taper_))
  {
    deg = FWIDiscStackPTR->deg;
    problem_fwd = std::move(problem_fwd_);
    initializeMeasureMask();
    Config::Get("UseGardener", useGardener);
  }


  [[nodiscard]] const Vector &GetWaveVec() const {
    if (FWIDiscStackPTR) return *FWIDiscStackPTR->wV;
    return *FWIDiscStack.wV;
  }

  [[nodiscard]] const Vector &GetMatVec() const {
    if (FWIDiscStackPTR) return *FWIDiscStackPTR->mV;

    return *FWIDiscStack.mV;
  }
  virtual std::unique_ptr<VAcousticMatFromVec> GetMaterial
      (const FWIVAcousticInversionProb &prob, const Vector &Mat) =0;
  virtual ~ForwardModel() = default;

  const FWIVAcousticInversionProb &GetProblem() { return *problem_fwd; }

  FWIVAcousticInversionProb &GetProblemWrite() { return *problem_fwd; }

  [[nodiscard]] virtual const char *Name() const = 0;

  bool UseBoxConstraint() { return Taper ? Taper->Conf.UseMinMaxTaper : false; }
  void ApplyBoxConstraint(FWIDataContainerDomain &iterateNonlinear,
                          FWIDataContainerDomain &iterateMUpdate) {
    Taper->ApplyMinMaxTaperAddition(iterateNonlinear.getMatWrite(), iterateMUpdate.getMatWrite(),
                                    *problem_fwd);
  }

  void Gardener(FWIDataContainerDomain &NonLinear) {
    Vector& NonLinearVec = NonLinear.getMatWrite();
    for (cell c = NonLinearVec.cells(); c != NonLinearVec.cells_end(); ++c) {
      double Vp = NonLinearVec(c(),1);
      if (Taper) {
        if (Taper->GetTaperVector()(c(), 0) < Eps)
          continue;
      }
      if(problem_fwd->getParametrization() == "taup"){
        NonLinearVec(c(),0) = 310.0 * pow(Vp,0.25);
      }else if(problem_fwd->getParametrization() == "taup_positive"){
        NonLinearVec(c(),0) = log(310) + 0.25*Vp;
      }else {
        Exit("CIAO");
      }
    }
  }

  virtual bool UseGardener(){ return useGardener;}

  [[nodiscard]] virtual std::string GetReconstructionType() const { return "Classical"; }

  virtual auto GetReconstructionVector() -> std::vector<unsigned int> {
    return recon.ReconVec;
  }

  [[nodiscard]] int GetParameterComponents() const { return problem_fwd->GetParamComps(); }

  virtual void NonlinearEvaluation(const FWIDataContainerDomain &nonlinearPoint,
                                   FWISubproblem &evaluation,
                                   std::vector<int> subproblem) = 0;

  virtual void AdjointLinearized(const FWIDataContainerDomain &nonlinearPoint,
                                 const FWISubproblem &, FWIDataContainerDomain &) = 0;

  virtual void Linearized(const FWIDataContainerDomain &nonlinearPoint,
                          const FWIDataContainerDomain &, FWISubproblem &) = 0;

  virtual void CalculateDefect(const std::vector<FWIDataContainerRange> &data_obs,
                               const FWISubproblem &nonlinearEvaluation,
                               FWISubproblem &defect,
                               std::vector<int> subproblem) = 0;

  virtual void MSDProcedure(const FWIDataContainerDomain &iterateNonlinear,
                            FWIDataContainerDomain &iterateUpdatePN,
                            FWIDataContainerDomain &iterateMUpdate,
                            FWISubproblem &defectLinearized) {
    Exit("notimplemented")
  }

  [[nodiscard]] double ScalarProductL2(const FWISubproblem &u,
                                               const FWISubproblem &v) const  {
    double skp = 0;
    for (size_t i = 0; i < u.size(); ++i) {
      skp += ScalarProductL2(u[i], v[i]);
    }
    return skp;
  }
  void Parameter2SeismogramMap(const Vector &m, DGSeismogram &s, int shot,
                                       Vectors *U_m) {
    problem_fwd->SetShot(shot);
    std::shared_ptr<DGAssemble> assemble = GetDGAssemble(m, *problem_fwd, *problem_fwd, true);
    FWITimeSolver TSolveR(GetWaveVec(), ABCH, TConf.rkorder);
    TSolveR.SolveAndMeasure(assemble, *Obs, s, U_m);
  }
  [[nodiscard]] virtual double ScalarProductL2(const FWIDataContainerRange &u,
                                               const FWIDataContainerRange &v) const = 0;

  [[nodiscard]] virtual double ScalarProductL2(const FWIDataContainerDomain &,
                                               const FWIDataContainerDomain &) const = 0;

  virtual void InitializePrecon(const Vector &M, const Vector &M_update,
                                const Vectors &U_M){}

  virtual void ApplyPrecon(const FWIDataContainerDomain &m,
                           FWIDataContainerDomain &) {}

  [[nodiscard]] int GetNumberOfShots() const {return problem_fwd->GetShotCount();}

  void createRealData(const FWIDataContainerDomain &m,
                      std::vector<FWIDataContainerRange> &obs){

    for (int shot = 0; shot < (*this).problem_fwd->GetShotCount(); ++shot) {
      mout << "Generating Real data for shot " << shot << endl;
      obs.emplace_back(DGSeismogram(Obs->GetSpaceOp().GetParaDat()));
      Parameter2SeismogramMap(m.getMat(), obs[shot].getSeismogramWrite(), shot, nullptr);
    }
  }
  std::shared_ptr<ProblemVAcoustic> GetWrappedProblem(
      const Vector &m, const FWIBndConditionsVAcoustic &bnd,
      const FWIRhsVAcoustic &rhs, bool fwd) {
    const auto &rhs_va = dynamic_cast<const FWIRhsVAcoustic &>(rhs);
    const auto &bnd_va = dynamic_cast<const FWIBndConditionsVAcoustic &>(bnd);
    auto MatI = GetMaterial(*problem_fwd,m);
    auto problem = std::make_shared<FWIProblemWrapperVA2>(rhs_va, bnd_va, std::move(MatI),
                                               problem_fwd->GetNumL(),
                                               problem_fwd->GetDim(),
                                               problem_fwd);
    return problem;
  }
  std::shared_ptr<DGAssemble> GetDGAssemble(const Vector &m, const FWIBndConditionsVAcoustic &bnd,
                                                    const FWIRhsVAcoustic &rhs,
                                                    bool fwd) {
    if (fwd) {
            auto assemble =
                std::make_shared<DGAssemble>(std::move(
                                                 GetWrappedProblem(m, bnd, rhs, fwd)),
                                                   true, 0, TConf.t0, TConf.T, TConf.dt);
      if (ABCH) {
              assemble->SetElementCache(ABCH->GetEleCache().get());
            } else {
              assemble->SetElementCache(eleCacheContainer->Get().get());
            }
            return assemble;
          } else {
            auto assemble =
                std::make_shared<DGVAcousticAssembleBwd>(std::move(
                    GetWrappedProblem(m, bnd,
                                      rhs, fwd)),
                                                         TConf.t0, TConf.T, TConf.dt);
            if (ABCH) {
              assemble->SetElementCache(ABCH->GetEleCache().get());
            } else {
              assemble->SetElementCache(eleCacheContainer->Get().get());
            }
            return assemble;
          }
    }
  void plot(const FWIDataContainerDomain &m, const std::string &name, int nr) {
      Plotter.Plot(m,name,nr);
    }

    void plot(const Vector &M, const std::string &name,int nr) {
      Plotter.plot(M,name,nr);
    };

    void plot(const Vectors &U, const std::string &name) {
      Plotter.plot(U, name);
    }
    const FWIObservationOperatorDGTS &GetObservationOperator() { return *Obs; }

    void fillSourceAndReceiverCellsFromProblem(Vector &M,
                                          const FWIVAcousticInversionProb &prob) const {
      for (cell c = M.cells(); c != M.cells_end(); ++c) {
        for (const auto &re : Obs->GetSpaceOp().GetSpecification()) {
          if (c.PointInCell(re)) { FillCellFromProblem(c, prob, M);
          }
        }
        for (int shot = 0; shot < problem_fwd->GetShotCount(); ++shot) {
          problem_fwd->SetShot(shot);
          int nrPointSrc = problem_fwd->getNrPointSrc();
          for (int i = 0; i < nrPointSrc; ++i) {
            Point src = problem_fwd->getPointSrcLoc(i);
            if (c.PointInCell(src)) { FillCellFromProblem(c, prob, M);
            }
          }
        }
      }
    }
    void SetPreconRegularization(double PRegu) { PreconRegu = PRegu; }

};



#endif
