#include "ForwardModelAAO.hpp"

#include <memory>

#include "AAO/AAOIntegralPrecon.hpp"
#include "AAO/FWILinearizedIntegralOperator.hpp"
#include "Assembler/FWILinearizedAssembler.hpp"
#include "BasicSolver.hpp"
#include "ForwardModel.hpp"
#include "Helper/FWIDataContainerDomain.hpp"
#include "Helper/FWIDataContainerRange.hpp"
#include "Helper/VectorsIntegrator.hpp"
#include "lib6_app/seismogram/ObservationSpecification.hpp"

RVector fillShapeValues(const DGAcousticElement &elem, const Point &receiverPoint) {
  RVector ret(elem.dimension());
  Point lz = elem.GetCell().GlobalToLocal(receiverPoint);
  int nP = elem.shape_size();
  int nV = elem.V_dimension();
  int p_dim = elem.p_dimension();
  int dim = elem.getDim();
  for (int i = 0; i < nP; ++i) {
    for (int l = 0; l < dim; ++l) {
      ret[i + l * nP] = elem.evaluateVelocityShape(lz, i)[l];
    }
  }
  for (int i = 0; i < nP; ++i) {
    double P_i = elem.evaluateShape(lz, i);
    for (int n = 0; n < p_dim; ++n) {
      ret[nV + n * nP + i] = P_i;
    }
  }
  return ret;
}

// void FWIForwardModelDGTSVAAllAtOnce::createRealData(const FWIDataContainerDomain &m,
//                                                     std::vector<FWIDataContainerRange> &obs) {
//   for (int shot = 0; shot < (*this).problem_fwd->GetShotCount(); ++shot) {
//     mout << "Generating 'real' data for shot " << shot << endl;
//     mout << "RAM in shot " << PPM->TotalMemoryUsage() << endl;
//     obs.emplace_back(FWISeismogram());
//     Parameter2SeismogramMap(m.getMat(), obs[shot].getSeismogramWrite(), shot, nullptr);
//     mout << "RAM before release " << PPM->TotalMemoryUsage() << endl;
//     mout << "RAM after release " << PPM->TotalMemoryUsage() << endl;
//
//   }
// }

void ForwardModelAAO::NonlinearEvaluation(const FWIDataContainerDomain &nonlinearPoint,
                                          FWISubproblem &evaluation, std::vector<int> subproblem) {
  if (subproblem.empty()) { Exit("subproblem not properly defined!") }
  for (size_t i = 0; i < subproblem.size(); ++i) {
    size_t shot = subproblem[i];
    problem_fwd->SetShot(shot);
    if (evaluation[i].getVectorsPointer() == nullptr) {
      evaluation[i].getVectorsPointer() =
          std::move(std::make_unique<Vectors>(TConf.timeSteps + 1, GetWaveVec()));
    }
    evaluation[i].getVectorsWrite() = 0.0;
    evaluation[i].getSeismogramWrite() = 0.0;
    Obs->MeasureFullWave(nonlinearPoint.getVectors(), evaluation[i].getSeismogramWrite());
    evaluateIntegralOperatorGInverse(nonlinearPoint.getVectors(), nonlinearPoint.getMat(),
                                     evaluation[i].getVectorsWrite(), true);
  }
}
void assembleRKRHS(const DGAssemble &assemble, const double dt, Vectors &RHSVectors, int order) {
  RVector c(0, 1);
  RVector b(0, 1);
  fillQuadrature(c, b, order);
  Vector tmp(0.0, RHSVectors[0]);
  for (size_t timeStep = 1; timeStep < RHSVectors.size(); ++timeStep) {
    Vector &RHS = RHSVectors[timeStep];
    for (size_t stage = 0; stage < c.size(); ++stage) {
      double evaluationTime = (timeStep - (1.0 - c[stage])) * dt;
      assemble.RHS(evaluationTime, tmp);
      RHS += (dt * b[stage]) * tmp;
    }
  }
}
void ForwardModelAAO::CalculateDefect(const std::vector<FWIDataContainerRange> &data_obs,
                                      const FWISubproblem &nonlinearEvaluation,
                                      FWISubproblem &defect, std::vector<int> subproblem) {
  mout << "before defectresize Defect " << PPM->TotalMemoryUsage() << endl;
  mout << "after Defect " << PPM->TotalMemoryUsage() << endl;
  defect.resize(0);
  for (size_t i = 0; i < subproblem.size(); ++i) {
    defect.emplace_back(DGSeismogram(Obs->GetSpaceOp().GetParaDat()),
                        Vectors(TConf.timeSteps + 1, GetWaveVec()));
    defect[i].getSeismogramWrite() = data_obs[subproblem[i]].getSeismogram();
    defect[i].getSeismogramWrite() -= nonlinearEvaluation[i].getSeismogram();
    Vectors RhsWF(TConf.timeSteps + 1, GetWaveVec());
    RhsWF.MakeAdditive();
    Vector Mat(0.0, GetMatVec());
    Mat.SetAccumulateFlag(false);
    FillVectorFromProblem(*problem_fwd, Mat);
    std::shared_ptr<DGAssemble> assemble(GetDGAssemble(Mat, *problem_fwd, *problem_fwd, true));
    assembleRKRHS(*assemble, TConf.dt, RhsWF, TConf.rkorder);
    defect[i].getVectorsWrite() = RhsWF;
    defect[i].getVectorsWrite() -= nonlinearEvaluation[i].getVectors();
  }
}

void ForwardModelAAO::evaluateIntegralOperator(const Vectors &NonlinearPointWave,
                                               const Vector &nonlinearPointMat,
                                               Vectors &WaveFieldEvaluation, inteOptions opt) {
  std::shared_ptr<RKOperator> RKOp(getRKOperator(nonlinearPointMat, NonlinearPointWave, opt));

  integralOperator inteOp(*RKOp);
  WaveFieldEvaluation.MakeAdditive();
  inteOp.run(NonlinearPointWave, WaveFieldEvaluation);
}

void ForwardModelAAO::evaluateIntegralOperatorGInverse(const Vectors &NonlinearPointWave,
                                                       const Vector &nonlinearPointMat,
                                                       Vectors &WaveFieldEvaluation,
                                                       inteOptions opt) {
  evaluateIntegralOperator(NonlinearPointWave, nonlinearPointMat, WaveFieldEvaluation, opt);
  (*GInv).ScalarProductInverse(WaveFieldEvaluation);
}

std::shared_ptr<RKOperator>
ForwardModelAAO::getRKOperator(const Vector &Mat, const Vectors &WaveFieldPoint, inteOptions opt) {
  std::shared_ptr<DGAssemble> assemble = GetDGAssemble(Mat, *problem_fwd, *problem_fwd, opt.fwd);
  return std::make_shared<RKOperator>(assemble, WaveFieldPoint[0], opt, TConf.rkorder, TConf.dt);
}
void ForwardModelAAO::evaluateLinearizedIntegralOperator(const Vectors &WaveFieldPoint,
                                                         const Vector &MatPoint,
                                                         const Vector &MatDirection,
                                                         Vectors &LinearizedEvaluation) {
  Vector MatDirection2(MatDirection);
  if (!MatDirection2.GetAccumulateFlag()) MatDirection2.Accumulate();
  FWILinearizedIntegralOperator linOp(problem_fwd, *eleCacheContainer->Get(), MatPoint,
                                      WaveFieldPoint, TConf.dt, TConf.rkorder);
  LinearizedEvaluation.Clear();
  linOp.Evaluate(MatDirection2, LinearizedEvaluation);
}

void ForwardModelAAO::evaluateAndAddLinearizedIntegralOperatorGInverse(
    const Vectors &WaveFieldPoint, const Vector &MatPoint, const Vector &MatDirection,
    Vectors &WaveFieldEvaluation) {
  Vectors LinearizedEvaluation(WaveFieldEvaluation);
  evaluateLinearizedIntegralOperator(WaveFieldPoint, MatPoint, MatDirection, LinearizedEvaluation);
  GInv->ScalarProductInverse(LinearizedEvaluation);
  WaveFieldEvaluation += LinearizedEvaluation;
}

void ForwardModelAAO::writeAdjointValuesInVector(const DGSeismogram &Seis, Vectors &U) const {
  const ObservationSpecification &spec = Obs->GetSpaceOp().GetSpecification();
  for (const auto &receiver : spec) {
    const Point &receiverPoint = receiver();
    if (!Obs->GetSpaceOp().GetSpaceSupport().contains(receiver)) continue;
    const std::list<Point> &midPoints = Obs->GetSpaceOp().GetSpaceSupport().at(receiver);
    const auto &tSeries = Seis.GetTimeSeries(receiverPoint);
    if (!midPoints.empty()) {
      for (const Point &midPoint : midPoints) {
        const DGAcousticElement &elem = *eleCacheContainer->Get()->at(midPoint);
        RVector shapeVals = fillShapeValues(elem, receiverPoint);
        int p_dim = elem.p_dimension();
        int nV = elem.V_dimension();
        int nP = elem.shape_size();
        for (int j = 0; j < U.size(); ++j) {
          FWIState meas(tSeries[j], 1);
          Mask->adjoint(meas);
          for (int k = 0; k < problem_fwd->GetDim(); ++k) {
            for (int i = 0; i < nP; ++i) {
              U[j](midPoint, i + k * nP) += meas[k] * shapeVals[i + k * nP];
            }
          }
          for (int i = 0; i < elem.shape_size(); ++i) {
            for (int n = 0; n < p_dim; ++n) {
              U[j](midPoint, nV + n * nP + i) +=
                  shapeVals[nV + n * nP + i] * meas[problem_fwd->GetDim() + n];
            }
          }
        }
      }
    }
  }
}

void ForwardModelAAO::AdjointMeasurement(const DGSeismogram &Seis,
                                         Vectors &AdjointLinearizedEvaluation) {
  if (Obs->GetSpaceOp().Name() == "point") {
    writeAdjointValuesInVector(Seis, AdjointLinearizedEvaluation);
    GInv->ScalarProductInverse(AdjointLinearizedEvaluation);
  }
}

void ForwardModelAAO::AdjointLinearized(const FWIDataContainerDomain &nonlinearPoint,
                                        const FWISubproblem &adjointLinearizedPoint,
                                        FWIDataContainerDomain &adjointLinearizedEvaluation) {

  Vectors &AdjointLinearizedEvaluation = adjointLinearizedEvaluation.getVectorsWrite();
  Vector &AdjointMat = adjointLinearizedEvaluation.getMatWrite();
  AdjointMat.Clear();
  AdjointLinearizedEvaluation.Clear();
  if (adjointLinearizedPoint.size() > 1) Exit("evaluate this again");
  for (const auto &i : adjointLinearizedPoint) {
    inteOptions opt(false, false, false);
    const Vectors &Lambda = i.getVectors();
    evaluateIntegralOperatorGInverse(Lambda, nonlinearPoint.getMat(), AdjointLinearizedEvaluation,
                                     opt);
    AdjointLinearizedEvaluation *= lambdaPenalty;
    AdjointMeasurement(i.getSeismogram(), AdjointLinearizedEvaluation);
    const Vectors &WaveFieldPoint = nonlinearPoint.getVectors();
    const Vector &Mat = nonlinearPoint.getMat();
    Vector MatTmpDiscrete(0.0, Mat);
    std::unique_ptr<Vectors> integratedVecs;
    calculateUpdateIntegral(WaveFieldPoint, Mat, Lambda, AdjointMat);
    AdjointMat *= lambdaPenalty;
    if (Taper) { Taper->ApplyTaperVector(adjointLinearizedEvaluation.getMatWrite()); }
  }
}

void ForwardModelAAO::Linearized(const Vectors &nonlinearPointWave, const Vector &nonlinearPointMat,
                                 const Vectors &linearizedWave, const Vector &linearizedMat,
                                 Vectors &linearizedEvaluationWave,
                                 DGSeismogram &linearizedEvaluationSeismo) {
  linearizedEvaluationWave.Clear();
  linearizedEvaluationSeismo = 0.0;
  inteOptions opt(true, false, false);
  evaluateIntegralOperatorGInverse(linearizedWave, nonlinearPointMat, linearizedEvaluationWave,
                                   opt);
  evaluateAndAddLinearizedIntegralOperatorGInverse(nonlinearPointWave, nonlinearPointMat,
                                                   linearizedMat, linearizedEvaluationWave);
  linearizedEvaluationWave *= lambdaPenalty;

  Obs->MeasureFullWave(linearizedWave, linearizedEvaluationSeismo);
}

void ForwardModelAAO::Linearized(const FWIDataContainerDomain &nonlinearPoint,
                                 const FWIDataContainerDomain &linearizedPoint,
                                 FWISubproblem &linearizedEvaluation) {
  if (linearizedEvaluation.size() > 1) Exit("Evaluate this again");
  for (auto &i : linearizedEvaluation) {
    Linearized(nonlinearPoint.getVectors(), nonlinearPoint.getMat(), linearizedPoint.getVectors(),
               linearizedPoint.getMat(), i.getVectorsWrite(), i.getSeismogramWrite());
  }
}

void ForwardModelAAO::calculateWeakFormVectors(Vectors &b) const {
  Vector tmp(b[0]);
  for (int i = 0; i < b.size(); ++i) {
    tmp = b[i];
    if (!tmp.GetAccumulateFlag()) tmp.SetAccumulateFlag(true);
    b[i] = GInv->ScalarProductMatrix() * tmp;
    // we dont need an accumulated vector for the multiplication since it is a block diagonal
    // matrix. So we set the flag true here. If the vector is already accumulated that's fine,too!
  }
}

double ForwardModelAAO::ScalarProductL2(const FWIDataContainerDomain &u,
                                        const FWIDataContainerDomain &v) const {
  const auto &U(u.getVectors());
  const auto &V(v.getVectors());
  const auto &UU(v.getMat());
  const auto &VV(u.getMat());
  double st = L2TScalarAAO(U, V);
  double s = nc::L2MaterialScalar(UU, VV, eleCacheContainer->Get(), recon.ReconVec);
  mout << endl;
  mout << " Spacetime: " << st << " Material: " << s << endl;
  mout << endl;
  return st + s;
}

auto ForwardModelAAO::L2tScalarAAO(const Vector &u, const Vector &v) const -> double {
  return u * (GInv->ScalarProductMatrix() * v);
}

auto ForwardModelAAO::L2TScalarAAO(const Vectors &u, const Vectors &v) const -> double {
  double l2t = 0.0;
  double weights[2] = {0.5, 0.5};
  for (int tsteps = 0; tsteps < u.size(); ++tsteps) {
    if (tsteps == 0 || tsteps == u.size() - 1) {
      l2t += weights[tsteps == 0 ? 0 : 1] * L2tScalarAAO(u[tsteps], v[tsteps]) * TConf.dt;
    } else {
      l2t += L2tScalarAAO(u[tsteps], v[tsteps]) * TConf.dt;
    }
  }
  return l2t;
}

double ForwardModelAAO::ScalarProductL2(const FWIDataContainerRange &u,
                                        const FWIDataContainerRange &v) const {
  const auto &U(u.getVectors());
  const auto &V(v.getVectors());
  double st = nc::L2TScalar(U, V, eleCacheContainer->Get(), TConf.dt);
  double s = u.getSeismogram() * v.getSeismogram();
  mout << endl;
  mout << " Spacetime: " << st << " Seismogram: " << s << endl;
  mout << endl;
  return st + s;
}

void ForwardModelAAO::calculateUpdateIntegral(const Vectors &WaveFieldPoint, const Vector &Mat,
                                              const Vectors &Lambda, Vector &AdjointMat) const {
  Vector MatTmp(Mat);
  VectorsIntegrator integrator(WaveFieldPoint, TConf.dt, TConf.rkorder);
  integrator.integrateOneStepAndGetVector();
  FWIAdjointCalculator adjointCalculator(Mat, problem_fwd->getParametrization(),
                                         *eleCacheContainer->Get(), recon, problem_fwd->GetNumL(),
                                         TConf.dt, problem_fwd);
  for (int timeStep = 1; timeStep < Lambda.size(); ++timeStep) {
    MatTmp.Clear();
    adjointCalculator.calculateL2AdjointOneStep(integrator.integrateOneStepAndGetVector(),
                                                WaveFieldPoint[timeStep], Lambda[timeStep], MatTmp);
    if (ABCHAAO.get()) { ABCHAAO->SetABCZero(MatTmp); }

    if (timeStep == Lambda.size() - 1) {
      AdjointMat += 0.5 * MatTmp;
    } else {
      AdjointMat += MatTmp;
    }
  }
}
DGSeismogram ForwardModelAAO::ApplyDataSpaceCG(const AAOIntegralPrecon &Prcon, DGSeismogram &r,
                                               Vectors &tmp1, Vectors &tmp2, int CGSteps) {
  DGSeismogram d(r);
  DGSeismogram z(r);
  DGSeismogram x(r);
  x = 0.0;
  z = 0.0;
  double beta{};
  double alpha{};
  double rnorm{};
  double residualnorm0{sqrt(r * r)};
  mout << "initial residual: " << residualnorm0 << endl;
  for (int step = 0; step < CGSteps; step++) {
    ApplyDataSpaceOp(Prcon, d, z, tmp1, tmp2);
    rnorm = r * r;
    alpha = rnorm / (d * z);
    x += alpha * d;
    r -= alpha * z;
    double residualnorm = sqrt(r * r);
    mout << "quotient to start: " << residualnorm / residualnorm0 << endl;
    beta = residualnorm / rnorm;
    d *= beta;
    d += r;
  }
  return x;
}

void ForwardModelAAO::ApplyDataSpaceOp(const AAOIntegralPrecon &Prcon, const DGSeismogram &seis,
                                       DGSeismogram &result, Vectors &tmp1, Vectors &tmp2) {
  result = 0.0;
  tmp1.Clear();
  AdjointMeasurement(seis, tmp1);
  tmp2 = tmp1 * Prcon;
  calculateWeakFormVectors(tmp2);
  tmp1 = Prcon * tmp2;
  Obs->MeasureFullWave(tmp1, result);
  result *= 1.0 / (lambdaPenalty * lambdaPenalty);
  result += seis;
}

void ForwardModelAAO::ApplyPrecon(const FWIDataContainerDomain &m,
                                  FWIDataContainerDomain &m_update) {
  const Vector &NonlinearPointMat = m.getMat();

  Vectors &UpdateWave = m_update.getVectorsWrite();
  bool usePenaltyForPrecon = true;
  Config::Get("usePenaltyForPrecon", usePenaltyForPrecon);
  int CGImprovement = 0;
  Config::Get("CGImprovement", CGImprovement);
  usePenaltyForPrecon = usePenaltyForPrecon || CGImprovement > 0;
  double AdjointNorm = 0.0;
  Vectors WaveTmp(UpdateWave);
  calculateWeakFormVectors(UpdateWave);
  AAOIntegralPrecon WavesolutionPrecon(GetWaveVec(), ABCHAAO.get(), *eleCacheContainer->Get(),
                                       NonlinearPointMat, *problem_fwd, TConf.t0, TConf.T,
                                       TConf.dt);
  WaveTmp = UpdateWave * WavesolutionPrecon;
  if (adaptivePenalty) {
    AdjointNorm = nc::L2TScalar(WaveTmp, WaveTmp, eleCacheContainer->Get(), TConf.dt);
  }
  calculateWeakFormVectors(WaveTmp);
  if (adaptivePenalty) {
    DGSeismogram tmpS(Obs->GetSpaceOp().GetParaDat());
    Obs->MeasureFullWave(WaveTmp, tmpS);
    lambdaPenalty = tmpS * tmpS / AdjointNorm * safetyFactor;
    mout << "Penalty Factor set to: " << lambdaPenalty << endl;
  }
  UpdateWave = WavesolutionPrecon * WaveTmp;
  if (usePenaltyForPrecon) UpdateWave *= 1.0 / (lambdaPenalty * lambdaPenalty);
  if (CGImprovement > 0) {
    DGSeismogram meas(Obs->GetSpaceOp().GetParaDat());
    Obs->MeasureFullWave(UpdateWave, meas);
    Vectors WaveTmp2(UpdateWave);
    WaveTmp2.Clear();
    AdjointMeasurement(meas, WaveTmp2);
    double Res{sqrt(L2TScalarAAO(WaveTmp2, WaveTmp2))};
    mout << " residual to  AWE second: " << Res << endl;
    WaveTmp2.Clear();
    DGSeismogram result =
        ApplyDataSpaceCG(WavesolutionPrecon, meas, WaveTmp, WaveTmp2, CGImprovement);
    WaveTmp2.Clear();
    AdjointMeasurement(result, WaveTmp2);
    WaveTmp = WaveTmp2 * WavesolutionPrecon;
    calculateWeakFormVectors(WaveTmp);
    WaveTmp2 = (1.0 / (lambdaPenalty * lambdaPenalty)) * (WavesolutionPrecon * WaveTmp);
    UpdateWave -= WaveTmp2;
    WaveTmp.Clear();
    evaluateIntegralOperatorGInverse(UpdateWave, NonlinearPointMat, WaveTmp, true);
    WaveTmp2.Clear();
    evaluateIntegralOperatorGInverse(WaveTmp, NonlinearPointMat, WaveTmp2, false);
    WaveTmp2 *= lambdaPenalty * lambdaPenalty;
    meas = 0.0;
    Obs->MeasureFullWave(UpdateWave, meas);
    WaveTmp.Clear();
    AdjointMeasurement(meas, WaveTmp);
    mout << " residual to AWE third: " << sqrt(L2TScalarAAO(WaveTmp, WaveTmp)) / Res << endl;
  }
}

void ForwardModelAAO::RestrictToVectorComp(Vector &u, int comp, int paraComps) const {
  for (int i = 0; i < paraComps; ++i) {
    if (i == comp) continue;
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
      u(c(), i) = 0.0;
    }
  }
}

void ForwardModelAAO::MSDProcedure(const FWIDataContainerDomain &iterateNonlinear,
                                   FWIDataContainerDomain &iterateUpdatePN,
                                   FWIDataContainerDomain &iterateMUpdate,
                                   FWISubproblem &defectLinearized) {
  const Vector &NonlinearPointMat = iterateNonlinear.getMat();
  const Vectors &NonlinearPointWave = iterateNonlinear.getVectors();
  Vector &UpdateMat = iterateUpdatePN.getMatWrite();
  Vectors &UpdateWave = iterateUpdatePN.getVectorsWrite();
  Vectors &defectLinearizedWaveField = defectLinearized[0].getVectorsWrite();
  DGSeismogram &defectLinearizedSeismo = defectLinearized[0].getSeismogramWrite();
  FWILinearizedAssembler fwiLinearizedAssembler(problem_fwd, *eleCacheContainer->Get(),
                                                NonlinearPointMat);
  DGSeismogram UpdateWaveSeismo(DGSeismogram(Obs->GetSpaceOp().GetParaDat()));
  Vector MTmp(NonlinearPointMat);
  MTmp.Clear();
  int Ninv = recon.ReconVec.size();
  std::vector<Vectors> Evals(Ninv, NonlinearPointWave);
  getLinearizedVectors(NonlinearPointMat, NonlinearPointWave, UpdateMat, Evals);
  inteOptions opt(true, false, false);
  Obs->MeasureFullWave(UpdateWave, UpdateWaveSeismo);
  Vector RHSUpperPart(UpdateMat);
  RHSUpperPart.Clear();
  Vector BRows(UpdateMat);
  BRows.Clear();
  MPC = std::make_unique<Matrix>(NonlinearPointMat);
  for (cell c = RHSUpperPart.cells(); c != RHSUpperPart.cells_end(); ++c) {
    auto &elem = *(eleCacheContainer->Get())->at(c());
    for (int param = 0; param < Ninv; ++param) {
      for (int i = 0; i < NonlinearPointWave.size(); ++i) {
        double factor = (i == 0 || i == NonlinearPointWave.size() - 1) ? 0.5 : 1.0;
        RHSUpperPart(c(), recon.ReconVec[param]) +=
            nc::CellWiseStepProduct(Evals[param][i], defectLinearizedWaveField[i], elem,
                                    factor * TConf.dt);
      }
    }
  }
  Vector tmp(UpdateWave[0]);
  Vector tmp2(UpdateWave[0]);

  tmp.Clear();
  std::shared_ptr<RKOperator> RKOp(getRKOperator(NonlinearPointMat, UpdateWave, opt));
  double EvalNormSqr = 0.0;
  double EvalDefectSKPSqr = 0.0;
  for (int i = 0; i < NonlinearPointWave.size(); ++i) {
    RKOp->loopFwdOneStep(UpdateWave, tmp, i);
    tmp2 = tmp;
    tmp2 *= lambdaPenalty;
    GInv->ScalarProductInverse(tmp2);
    double factor = (i == 0 || i == NonlinearPointWave.size() - 1) ? 0.5 : 1.0;
    for (cell c = BRows.cells(); c != BRows.cells_end(); ++c) {
      auto &elem = *(eleCacheContainer->Get())->at(c());
      for (int param = 0; param < Ninv; ++param) {
        BRows(c(), recon.ReconVec[param]) +=
            nc::CellWiseStepProduct(Evals[param][i], tmp2, elem, factor * TConf.dt);
      }
    }
    EvalNormSqr += factor * nc::L2tScalar(tmp2, tmp2, eleCacheContainer->Get(), TConf.dt);
    EvalDefectSKPSqr += factor * nc::L2tScalar(tmp2, defectLinearizedWaveField[i],
                                               eleCacheContainer->Get(), TConf.dt);
  }

  assembleAAOInv(Evals, *MPC);
  Vector RHSUpperPartSol(RHSUpperPart);
  double DiagonalValue = EvalNormSqr + UpdateWaveSeismo * UpdateWaveSeismo;
  double LowerRHS = EvalDefectSKPSqr + UpdateWaveSeismo * defectLinearizedSeismo;
  MOUT(DiagonalValue);
  MOUT(BRows.norm());
  RHSUpperPart.SetAccumulateFlag(true);
  double valLower = BRows * ((*MPC) * RHSUpperPart);
  BRows.SetAccumulateFlag(true);
  double valDiagonal = BRows * ((*MPC) * BRows);
  BRows.SetAccumulateFlag(false);
  LowerRHS -= valLower;
  DiagonalValue -= valDiagonal;
  MOUT(DiagonalValue);
  double Sol = LowerRHS / DiagonalValue;
  MOUT(Sol);
  BRows *= Sol;
  RHSUpperPart.MakeAdditive();
  RHSUpperPart -= BRows;
  RHSUpperPart.SetAccumulateFlag(true);
  BRows = (*MPC) * RHSUpperPart;
  for (cell c = RHSUpperPart.cells(); c != RHSUpperPart.cells_end(); ++c) {
    for (int i = 0; i < Ninv; ++i) {
      UpdateMat(c(), recon.ReconVec[i]) *= BRows(c(), recon.ReconVec[i]);
    }
  }
  UpdateWave *= Sol;
}

void ForwardModelAAO::assembleAAOInv(const std::vector<Vectors> &Evals, Matrix &PHess) const {
  int Ninv = recon.ReconVec.size();
  if (Ninv > 2)
    Exit("find other inversion method e.g. use schur complement in addition to this") PHess = 0.0;
  double maxA = -infty;
  double minA = infty;
  for (const auto &elemIt : *eleCacheContainer->Get()) {
    const auto &elem = *elemIt.second;
    const Cell &C = elemIt.second->GetCell();
    DGRowEntries PHess_c(PHess, C, C);

    if (Ninv == 1) {
      double A = nc::CellWiseTimeProduct(Evals[0], Evals[0], elem, TConf.dt);
      maxA = std::max(A, maxA);
      minA = A > 1e-100 ? std::min(A, minA) : minA;
      PHess_c(recon.ReconVec[0], recon.ReconVec[0]) = A;
    } else if (Ninv == 2) {
      double D_11 = nc::CellWiseTimeProduct(Evals[0], Evals[0], elem, TConf.dt);
      double D_12 = nc::CellWiseTimeProduct(Evals[1], Evals[0], elem, TConf.dt);
      double D_22 = nc::CellWiseTimeProduct(Evals[1], Evals[1], elem, TConf.dt);
      double Tmp = (D_22 - D_12 * D_12 / D_11);
      double Epsloc = 1e-100;
      if (Tmp < Epsloc || D_11 < Epsloc) {
        PHess_c(recon.ReconVec[0], recon.ReconVec[0]) = 0.0;
        PHess_c(recon.ReconVec[1], recon.ReconVec[0]) = 0.0;
        PHess_c(recon.ReconVec[0], recon.ReconVec[1]) = 0.0;
      } else {
        double Tmp2 = 1.0 / Tmp;
        PHess_c(recon.ReconVec[0], recon.ReconVec[0]) =
            1.0 / D_22 + (D_12 * D_12) / (D_22 * D_22) * Tmp2;
        PHess_c(recon.ReconVec[1], recon.ReconVec[0]) = -Tmp2 * D_12 / D_11;
        PHess_c(recon.ReconVec[0], recon.ReconVec[1]) = -Tmp2 * D_12 / D_11;
        PHess_c(recon.ReconVec[1], recon.ReconVec[1]) = Tmp2;
      }
    }
  }
  maxA = PPM->Max(maxA);
  minA = PPM->Min(minA);
  if (abs(minA - infty) < 1e-100) { minA = 0.0; }
  mout << "Maximal Value in Diagonal Matrix: " << maxA << endl;
  mout << "Minimal Value in Diagonal Matrix: " << minA << endl;
  if (Ninv == 1) {
    if (minA != maxA) {
      for (const auto &elemIt : *eleCacheContainer->Get()) {
        const Cell &C = elemIt.second->GetCell();
        DGRowEntries PHess_c(PHess, C, C);
        double Epsloc = 1e-100;
        double A = PHess_c(recon.ReconVec[0], recon.ReconVec[0]);
        if (A < Epsloc) {
          PHess_c(recon.ReconVec[0], recon.ReconVec[0]) = 0.0;
        } else {
          A += AAOInvRegularizationValue > 0 ? AAOInvRegularizationValue : 0.0;
          PHess_c(recon.ReconVec[0], recon.ReconVec[0]) = 1.0 / A;
        }
      }
    }
  }
  (*MPC).CreateSparse();
}

void ForwardModelAAO::getLinearizedVectors(const Vector &m, const Vectors &u_fd,
                                           const Vector &Direc, std::vector<Vectors> &Evals) {
  FWILinearizedAssembler fwiLinearizedAssembler(problem_fwd, *eleCacheContainer->Get(), m);
  int Ninv = recon.ReconVec.size();
  if (Ninv > 2)
    Exit("find other inversion method e.g. use schur complement in addition to this") for (
        int param = 0; param < Ninv; param++) {
      VectorsIntegrator inte(u_fd, TConf.dt, TConf.rkorder);
      Vector matTmp(Direc);
      RestrictToVectorComp(matTmp, recon.ReconVec[param], problem_fwd->GetParamComps());
      for (int i = 0; i < u_fd.size(); ++i) {
        fwiLinearizedAssembler.MPrimeMatVecDiagonal(matTmp, inte.integrateOneStepAndGetVector(),
                                                    Evals[param][i]);
      }
      Evals[param] *= lambdaPenalty;
    }
}

void ForwardModelAAO::InitializePrecon(const Vector &M, const Vector &M_update,
                                       const Vectors &U_M) {}
