#include "ForwardModelRed.hpp"

#include <memory>

#include "BasicSolver.hpp"
#include "FWIProblemWrapper.hpp"
#include "Helper/FWIDataContainerDomain.hpp"
#include "Helper/FWIDataContainerRange.hpp"
#include "Helper/VectorFillFunctions.hpp"
#include "PseudoHessianPreconditioner.hpp"

double FWIForwardModelRed::L2TScalar(const Vectors &u, const Vectors &v) const {
  double l2t = 0.0;
  for (int tsteps = 0; tsteps < u.size(); ++tsteps) {
    if (tsteps == 0 || tsteps == u.size() - 1) {
      l2t += 0.5 * L2tScalar(u[tsteps], v[tsteps]);
    } else {
      l2t += L2tScalar(u[tsteps], v[tsteps]);
    }
  }
  return l2t * TConf.dt;
}



double FWIForwardModelRed::L2tScalar(const Vector &u, const Vector &v) const {
  double e = 0;
  for (const auto &elem: *eleCacheContainer->Get()) {
    e += nc::CellWiseStepProduct(u, v, *elem.second, 1.0);
  }
  return PPM->Sum(e);
}

double FWIForwardModelRed::L2Scalar(const Vector &u,
                                    const Vector &v) const { // needs to change if not constant materials are used.
  double l2 = 0.0;
  for (const auto &elemIt: *eleCacheContainer->Get()) {
    double a = elemIt.second->Area();
    row ru = u.find_row(elemIt.first);
    row rv = v.find_row(elemIt.first);
    for (auto k: recon.ReconVec) {
      double para_u = u(ru, k);
      double para_v = v(rv, k);
      l2 += para_u * para_v * a;
    }
  }
  return PPM->Sum(l2);
}

int FWIForwardModelRed::GetDim() const { return problem_fwd->GetDim(); }

FWIForwardModelRed::~FWIForwardModelRed() = default;
void FWIForwardModelRed::NonlinearEvaluation(
    const FWIDataContainerDomain &nonlinearPoint, FWISubproblem &evaluation,
    std::vector<int> subproblem) {
  if (subproblem.empty()) {
    Exit("subproblem not properly defined!")
  }
  if (uM.size() != subproblem.size()) {
    if (uM.size() > subproblem.size()) {
      for (size_t i = subproblem.size(); i < uM.size(); ++i)
        uM[i].release();
    }
    if (uM.size() < subproblem.size()) {
      for (size_t i = uM.size(); i < subproblem.size(); ++i)
        uM.push_back(
            std::make_unique<Vectors>(TConf.timeSteps + 1, GetWaveVec()));
    }
  }
  if (subproblem.size() > 1)
  Exit("Not like this") for (size_t i = 0; i < subproblem.size(); ++i) {
    mout << "shot: " << subproblem[i] << ": Solving forward problem..."
         << endl;
    Nonlinear(nonlinearPoint.getMat(), *uM[i],
              evaluation[i].getSeismogramWrite(), subproblem[i]);
    mout << "Assembling Linearized Operators" << endl;
    MakeLinearAndAdjointOperator(nonlinearPoint.getMat(), *uM[i]);
    mout << "Done!" << endl;
  }
}
void FWIForwardModelRed::InitializePrecon(const Vector &M,
                                          const Vector &M_update,
                                          const Vectors &U_M) {
  if (precon) {
    if (!PHess)
      PHess = std::make_unique<PseudoHessianPreconditioner>(problem_fwd,
                                                            *eleCacheContainer,
                                                            recon.ReconVec, recon,
                                                            TConf.dt);
    PHess->assemble(M, M_update, *uM[0], PreconRegu);
  }
}
void FWIForwardModelRed::AdjointLinearized(
    const FWIDataContainerDomain &nonlinearPoint,
    const FWISubproblem &adjointLinearizedPoint,
    FWIDataContainerDomain &adjointLinearizedEvaluation) {
  FWIDataContainerDomain tmp(GetReconstructionType(), GetMatVec(), GetWaveVec(),
                             TConf.timeSteps);
  for (size_t i = 0; i < uM.size(); ++i) {
    AdjOp->CalculateAdjoint(adjointLinearizedPoint[i].getSeismogram(),
                            tmp.getMatWrite(), nullptr);
    adjointLinearizedEvaluation.getMatWrite() += tmp.getMat();
  }
  if (Taper) {
    Taper->ApplyTaperVector(adjointLinearizedEvaluation.getMatWrite());
  }

  adjointLinearizedEvaluation.getMatWrite().MakeAdditive();
}
void FWIForwardModelRed::Linearized(
    const FWIDataContainerDomain &nonlinearPoint,
    const FWIDataContainerDomain &linearizedPoint,
    FWISubproblem &linearizedEvaluation) {
  for (size_t i = 0; i < uM.size(); ++i) {
    LinOp->SolveAndMeasure(linearizedPoint.getMat(), *Obs,
                           linearizedEvaluation[i].getSeismogramWrite(),
                           nullptr);
  }
}
void FWIForwardModelRed::CalculateDefect(
    const std::vector<FWIDataContainerRange> &data_obs,
    const FWISubproblem &nonlinearEvaluation, FWISubproblem &defect,
    std::vector<int> subproblem) {
  if (subproblem.size() > 1){
    Exit("Check this for bigger subproblems!")}
  defect.resize(0);
  for (size_t i = 0;i < subproblem.size();++i) {
    defect.emplace_back(DGSeismogram{Obs->GetSpaceOp().GetParaDat()},
                        Vectors{TConf.timeSteps,GetWaveVec()});
    defect[i].getSeismogramWrite() = data_obs[subproblem[i]].getSeismogram();
    defect[i].getSeismogramWrite() -= nonlinearEvaluation[i].getSeismogram();
  }
}
std::unique_ptr<VAcousticMatFromVec>
FWIForwardModelRed::GetMaterial(const FWIVAcousticInversionProb &prob, const Vector &Mat){
  return std::make_unique<VAcousticMatFromVec>(prob,Mat,
                                               ABCH.get(), nullptr);
}