#ifndef FWI_FWISOURCES_HPP
#define FWI_FWISOURCES_HPP

#include <fstream>
#include <functional>

#include "../../../../DSPFilters/Dsp.h"
#include "CVector.hpp"
#include "Helper/FWIHelpers.hpp"
#include "spline.h"
class SourceWaveletBase {
protected:
  double dt = 0.0;
  RVector evalsOnGrid;
public:
  void SetEvalsOnGrid(const RVector &evalsOnGrid_) ;

public:
  [[nodiscard]] const RVector &GetEvalsOnGrid() const;

protected:
  RVector timeGrid;
  tk::spline SplineRepresent;
  double signalFreq=-1.0;
public:
  SourceWaveletBase() = default;

  SourceWaveletBase(const SourceWaveletBase& wav){
    dt = wav.dt;
    evalsOnGrid = wav.evalsOnGrid;
    timeGrid = wav.timeGrid;
    SplineRepresent =wav.SplineRepresent;
    signalFreq=wav.signalFreq;

  }
  void Bandpass(double centerFreq, double width, int order);

  void BandpassIFOS(double,double);

  void Lowpass(double freq, int order);

  [[nodiscard]] CVector GetFFT() const;

  double operator()  (double t) const;

  [[nodiscard]] virtual std::string GetName() const{ return "";};

  void WriteToFile(const std::string &fn) const;
  [[nodiscard]] double GetSignalFreq() const {
    if(signalFreq == -1){
      Exit("No Dominant Frequency set.")
    }
    return signalFreq;
  }
  virtual ~SourceWaveletBase()  = default;
};

class SourceFromFile : public SourceWaveletBase {
public:
  explicit SourceFromFile(const std::string &fn,tk::spline::spline_type type = tk::spline::cspline ) {
    json j;
    std::ifstream ifs(fn + ".json");
    if (!ifs.good())
      Exit("Failed to open file" + fn);
    nlohmann::json SourceWavelet = nlohmann::json::parse(ifs);
    for (const auto &t : SourceWavelet.at("Times")){
      timeGrid.push_back(t);
    }
    for (const auto &d : SourceWavelet.at("Data")){
    evalsOnGrid.push_back(d);
  }
  SplineRepresent = tk::spline();
  SplineRepresent.set_boundary(tk::spline::second_deriv,0.0,tk::spline::second_deriv,0.0);
  SplineRepresent.set_points(timeGrid.asVector(),evalsOnGrid.asVector(),type);
  }
  [[nodiscard]] std::string GetName()const override { return "SourceFromFile"; }
};

class SourceWaveletFromF : public SourceWaveletBase {
  std::vector<double> evalPoints;
public:
  SourceWaveletFromF(const std::function<double(double)> &, double, double, double, int);

  SourceWaveletFromF(const SourceWaveletFromF& wav) : SourceWaveletBase(wav) {
    dt = wav.dt;
    evalsOnGrid = wav.evalsOnGrid;
    timeGrid = wav.timeGrid;
    SplineRepresent =wav.SplineRepresent;
    signalFreq=wav.signalFreq;
    evalPoints = wav.evalPoints;
  }

};

class RickerDerivativeWavelet : public SourceWaveletFromF {
public:

  RickerDerivativeWavelet(double signalFreq, double delay, double cutoff, double dt, double T,
                          int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double, double)>(RickerDerivativeDelay),
                    signalFreq,
                    std::placeholders::_1, delay, cutoff), dt, T, signalFreq, rkorder) {
  }
  [[nodiscard]] std::string GetName() const override { return "RickerDerivative"; }
};

class RickerWavelet : public SourceWaveletFromF {
public:
  RickerWavelet(double signalFreq, double delay, double cutoff, double dt, double T, int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double, double)>(RickerDelay), signalFreq,
                    std::placeholders::_1, delay, cutoff), dt, T, signalFreq, rkorder) {
  }
  [[nodiscard]] std::string GetName() const override { return "Ricker"; }
};
double sinDelayIntegral(double signalFreq, double t, double delay,double internFac);
class SinWaveletIntegral : public SourceWaveletFromF {
public:
  SinWaveletIntegral(double signalFreq, double delay, double internFac,double dt, double T, int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double,double)>(sinDelayIntegral), signalFreq,
                    std::placeholders::_1, delay,internFac), dt, T, signalFreq, rkorder) {
  }
  [[nodiscard]] std::string GetName() const override { return "SinIntegral"; }
};
double sinDelay(double signalFreq,double t, double delay,double internFac);

class SinWavelet : public SourceWaveletFromF {
public:
  SinWavelet(double signalFreq, double delay, double internFac,double dt, double T, int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double,double)>(sinDelay), signalFreq,
                    std::placeholders::_1, delay,internFac), dt, T, signalFreq, rkorder) {
  }
  [[nodiscard]] std::string GetName() const override { return "Sin"; }
};
double CosDelay(double signalFreq,double t, double delay,double internFac);
class CosWavelet : public SourceWaveletFromF {
public:
  CosWavelet(double signalFreq, double delay, double internFac,double dt, double T, int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double,double)>(CosDelay), signalFreq,
                    std::placeholders::_1, delay,internFac), dt, T, signalFreq, rkorder) {
  }
  [[nodiscard]] std::string GetName() const override { return "Cos"; }
};
std::unique_ptr<SourceWaveletFromF> GetSourceWavelet(const TimeConf &TConf, const SourceConf &SConf);

#endif //FWI_FWISOURCES_HPP
