//
// Created by cr on 07/03/24.
//

#include "FWISeismogramTest.hpp"

#include <memory>

#include <gtest/gtest.h>

#include "DGViscoAcousticProblem.hpp"
#include "MeshesCreator.hpp"
#include "Observation/DGSeismogram.hpp"
#include "Observation/FWIObservationOperatorDGTS.hpp"
#include "Observation/ObservationCreator.hpp"
#include "TestEnvironment.hpp"
#include "Vector.hpp"
class MiniProb {
public:
  int dim;
  int numL;
  MiniProb() = delete;
  MiniProb(int dim_, int numL_) : dim(dim_), numL(numL_) {}
  virtual double Ut(double t, const Point &NodalPoint, int d) const = 0;
  double MeasureVal(double t, const Point &P) const {
    double returnValue = 0.0;
    for (int n = 0; n < numL + 1; ++n) {
      returnValue += Ut(t, P, 2 + n);
    }
    return returnValue;
  }
};
class MiniProbOne : public MiniProb {
public:
  MiniProbOne() : MiniProb(2, 2) {}
  double Ut(double t, const Point &Location, int d) const override {
    switch (d) {
      case 0:
        return Location[0] + Location[1] + t * t;
      case 1:
        return (Location[0] + Location[1] + t * t) *
               (Location[0] + Location[1] + t * t);
      case 2:
        return Location[0] * t + Location[1] * Location[0] * t * t;
      case 3:
        return Location[1] * Location[0];
      case 4:
        return Location[0] * Location[0] + t + t * t;
      default:
        THROW("Wrong Call");
    }
  }
};
class MiniProbConstant : public MiniProb {
public:
  MiniProbConstant() : MiniProb(2, 2) {}
  double Ut(double t, const Point &Location, int d) const override {
    switch (d) {
      case 0:
        return Location[0] + Location[1] + t * t;
      case 1:
        return (Location[0] + Location[1] + t * t) *
               (Location[0] + Location[1] + t * t);
      case 2:
        return 1;
      case 3:
        return 2;
      case 4:
        return 3;
      default:
        THROW("Wrong Call");
    }
  }
};
void createSolutionVector(const MiniProb &Problem, Vector &Vec, double t) {
  Vec = 0.0;
  for (cell c = Vec.cells(); c != Vec.cells_end(); ++c) {
    DGAcousticElement elem(Vec, *c, Problem.numL);
    for (int d = 0; d < elem.u_dimension(); ++d) {
      for (int nodalPointID = 0; nodalPointID < elem.NodalPoints();
           ++nodalPointID) {
        const auto &NodalPoint = elem.NodalPoint(nodalPointID);
        Vec(c(), d * elem.shape_size() + nodalPointID) =
            Problem.Ut(t, NodalPoint, d);
      }
    }
  }
}
struct TestParams {
  std::string problem;
};

class FWISeismogramTest : public TestWithParam<TestParams> {
protected:
  std::unique_ptr<DGSeismogram> FwdSeis;
  std::shared_ptr<DGDiscretization> Discretization;
  std::unique_ptr<Vector> Vec;
  MiniProbOne Problem;
  std::unique_ptr<FWIObservationOperatorDGTS> Obs;
  std::unique_ptr<EleCacheContainer> eleCache;
  std::shared_ptr<Meshes> meshes;

public:
  FWISeismogramTest() {
    double LeftBnd = 0.0;
    double RightBnd = 1.0;
    double BottomBnd = 0.0;
    double TopBnd = 1.0;
    int dimX = 33;
    int dimY = 33;
    int deg = 2;
    std::vector<int> bnds = {0, 0, 0, 0};
    auto geo =
        std::make_shared<RectangularGeometry2D>(std::pair<double, double>(
                                                    {LeftBnd, RightBnd}),
                                                std::pair<double, double>(
                                                    {BottomBnd, TopBnd}),
                                                dimX, dimY, bnds, "TestMesh");
    meshes = MeshesCreator("RectangularFWI")
                 .WithCoarseGeometry(geo)
                 .WithDistribute("RCB")
                 .CreateShared();
    int stateComps = Problem.dim + 1 + Problem.numL;
    Discretization =
        std::make_unique<DGDiscretization>(*meshes, deg, stateComps);
    Vec = std::make_unique<Vector>(Discretization);
    eleCache = std::make_unique<EleCacheContainer>(*Vec, Problem.numL);
    std::vector<Point> recpts = {{0.1, 0.1},
                                 {0.9, 0.1},
                                 {0.9, 0.9},
                                 {0.1, 0.9}};
    auto spec = ObservationSpecificationBuilder()
                    .WithT(1.0)
                    .Withdt(0.01)
                    .Witht0(0)
                    .WithReceivers(recpts)
                    .CreateShared();
    Obs = ObservationCreator(eleCache->Get(), Vec->GetMesh(), Problem.numL)
              .WithSpec(spec)
              .CreateUnique();
  }
};

TEST_F(FWISeismogramTest, MeasurePressureTest) {
  const auto &spec = Obs->GetSpaceOp().GetSpecification();
  const auto &specTimesteps =
      Obs->GetSpaceOp().GetSpecification().GetNumberOfTimeSteps();
  DGSeismogram seismo(Obs->GetSpaceOp().GetParaDat());
  for (int timestep = 0; timestep < specTimesteps; ++timestep) {
    createSolutionVector(Problem, *Vec, spec.TimeAtStep(timestep));
    Obs->MeasureSingleWaveField(*Vec, seismo, timestep);
  }
  std::vector<double> errors;
  for (const auto &receiverLocation :
       Obs->GetSpaceOp().GetParaDat().LocalReceivers) {
    const auto &receiverTimeSeries = seismo.GetTimeSeries(receiverLocation);
    double error = 0.0;
    double energy = 0.0;
    for (int receiverTimeStep = 0; receiverTimeStep < specTimesteps;
         ++receiverTimeStep) {
      double t = spec.TimeAtStep(receiverTimeStep);
      double errorval = receiverTimeSeries[receiverTimeStep] -
                        Problem.MeasureVal(t, receiverLocation);
      error += errorval * errorval;
      energy += receiverTimeSeries[receiverTimeStep] *
                receiverTimeSeries[receiverTimeStep];
    }
    errors.push_back(sqrt(error / energy));
  }
  EXPECT_VECTOR_NEAR(errors, std::vector<double>(errors.size(), 0.0), 1e-15);
}
TEST_F(FWISeismogramTest, ReadAndWriteTest) {
  const auto &spec = Obs->GetSpaceOp().GetSpecification();
  const auto &specTimesteps =
      Obs->GetSpaceOp().GetSpecification().GetNumberOfTimeSteps();
  DGSeismogram seismo(Obs->GetSpaceOp().GetParaDat());
  for (int timestep = 0; timestep < specTimesteps; ++timestep) {
    createSolutionVector(Problem, *Vec, spec.TimeAtStep(timestep));
    Obs->MeasureSingleWaveField(*Vec, seismo, timestep);
  }
  seismo.WriteToFile("SeismogramTmp");
  DGSeismogram seismo2(Obs->GetSpaceOp().GetParaDat());
  seismo2.ReadFromFile("SeismogramTmp");
  seismo2 -= seismo;
  EXPECT_NEAR(sqrt((seismo2 * seismo2) / (seismo * seismo)), 0.0, 1e-50);
}

int main(int argc, char **argv) {
  return MppTest(MppTestBuilder(argc, argv)
                     .WithConfigEntry("default", 0)
                     .WithScreenLogging()
                     .WithConfPath("../")
                     .WithGeoPath("../conf/geo/")
                     .WithPlotPath("data/vtu/")
                     .WithPPM())
      .RUN_ALL_MPP_TESTS();
}