//
// Created by cr on 19.11.22.
//

#include "FWITestLinearized.hpp"

#include <experimental/random>

#include "FWIProblemWrapper.hpp"
#include "Helper/FWIHelpers.hpp"
#include "Helper/VectorFillFunctions.hpp"
#include "MeshIndex.hpp"
#include "Observation/ObservationCreator.hpp"
#include "Observation/DGSeismogram.hpp"
#include "Observation/FWIObservationOperatorDGTS.hpp"
#include "Problems/FWIInversionProblems.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "TestEnvironment.hpp"
#include "lib6_app/seismogram/ObservationSpecification.hpp"

std::map<Point, const Cell&> GetOrderedCells(cell start, cell end){
  std::map<Point, const Cell&> result;
  for(cell c=start; c!= end; c++){
    result.emplace(c(), *c);
  }
  return result;
}

struct TestParams {
  std::string model;
  std::string parametrization;
  int param;
  bool Measure;
  std::vector<double> values = std::vector<double>(6, 0.0);

  friend void PrintTo(const TestParams& param, std::ostream* os) {
    *os << "Model: " << param.model << " paramet: " <<  param.parametrization
        << " param: " << param.param << " measure: " << param.Measure
        << " values "
        << param.values[0] << ","
        << param.values[1] << ","
        << param.values[2] << ","
        << param.values[3] << ","
        << param.values[4] << ","
        << param.values[5] << endl;
  }

};

class FWITestLinearized : public TestWithParam<TestParams> {
protected:
  std::unordered_map<Point,double> SeisMNorms;
  std::unique_ptr<WPConf> Conf = nullptr;
  std::shared_ptr<Meshes> meshes;
  std::shared_ptr<FWIVAcousticInversionProb> prob;
  std::unique_ptr<FWIDiscretizationStack> discStack;
  std::unique_ptr<EleCacheContainer> eleCache;
  std::shared_ptr<ABCHandler> PMLH;
  std::unique_ptr<DGVAcousticPlotter> plotter;
  std::unique_ptr<FWIObservationOperatorDGTS> Obs;
  std::unique_ptr<DGSeismogram> s_m;
  std::unique_ptr<Vector> Mat;
  std::unique_ptr<Vectors> U_m;
  std::unique_ptr<FWILinOpRed> LinOp;
  const std::vector<unsigned int> reconlist{0, 1, 2};
public:
  FWITestLinearized() {
    Config::Initialize(TestFWILinearized);
    Conf = std::make_unique<WPConf>();
    meshes = getFWIMesh();
    prob = GetInversionProblem(*Conf);
    prob->setParametrization(GetParam().parametrization);
    int deg = 0;
    Config::Get("deg", deg);
    discStack = std::make_unique<FWIDiscretizationStack>(meshes, prob->GetParamComps(),
                                                         prob->GetStateComponents(),
                                                         deg, MeshIndex{Conf->level, -1});
    eleCache = std::make_unique<EleCacheContainer>(*discStack->wV, prob->GetNumL());
    plotter = std::make_unique<DGVAcousticPlotter>(prob->GetNumL(), eleCache->Get().get());
    std::shared_ptr<ObservationSpecification> spec = ObservationSpecificationBuilder().Witht0(Conf->TConf.t0)
        .Withdt(Conf->TConf.dt).WithT(Conf->TConf.T).CreateShared();
    auto mask = GetMeasureMask(meshes->dim(), prob->GetNumL(), 0);
    Obs = ObservationCreator(eleCache->Get(), discStack->GetMesh(),
                                prob->GetNumL()).CreateUnique();
    PMLH = std::shared_ptr<ABCHandler>(nullptr);
    Mat = std::make_unique<Vector>(*discStack->mV);
    Mat->Clear();
    FillVectorFromProblem(*prob, *Mat);
    U_m = std::make_unique<Vectors>(Conf->TConf.timeSteps + 1, *discStack->wV);
    s_m = std::make_unique<DGSeismogram>(Obs->GetSpaceOp().GetParaDat());
    prob->SetShot(0);
    NonlinearEvaluation(*Mat,*U_m,*s_m);
    MakeLinearOp();

  }
  void MakeLinearOp(){
    if(GetParam().model == "L2") {
      LinOp = std::make_unique<FWILinOpRed>(*U_m, *Mat, prob, PMLH, Conf->TConf, eleCache->Get(),
                                            true);
    }else{
      Exit("not implemented")
    }
  }
  void NonlinearEvaluation(const Vector& Mat_,Vectors &uM, DGSeismogram& sM) {
    FWITimeSolver TSolveR(*discStack->wV, PMLH, Conf->TConf.rkorder);
    TSolveR.SolveAndMeasure(GetAssemble(Mat_), *Obs, sM, &uM);
    if(GetParam().model == "L2"){
    }else{
      Exit("not implemented")
    }
  }
  std::shared_ptr<DGAssemble> GetAssemble(const Vector &Material) {
    auto Mp = std::make_unique<Vector>(*discStack->mV);
    Mp->Clear();
    ReparametrizeMaterialVector(Material, *prob, *Mp);
    Mp->Accumulate();
    auto WrappedProblem = std::make_shared<FWIProblemWrapper>(*prob, *prob, std::move(Mp),
                                              prob->GetNumL(), prob->GetDim(),
                                                                prob);
    auto assemble = std::make_shared<DGAssemble>(WrappedProblem, true, -1, Conf->TConf.t0,
                                                 Conf->TConf.T,Conf->TConf.dt);
    assemble->SetElementCache(eleCache->Get().get());
    return std::move(assemble);
  }

  void TearDown() override {
    Config::Close();
  }
};

TEST_P(FWITestLinearized, FWITestLinearized) {
  Vector Rand(*Mat);
  Rand.Clear();
  auto srcLoc = prob->getPointSrcLoc(0);
  std::experimental::reseed(1);
  for (auto &[point, c]: GetOrderedCells(Rand.cells(), Rand.cells_end())) {
    if (c.PointInCell(srcLoc)) continue;
    auto r = double(std::experimental::randint(0, 10000) / 10000.0);
    Rand(point, GetParam().param) = r;
  }
  Rand *= 1.0 / nc::L2MaterialNorm(Rand, eleCache->Get(), reconlist);
  Vector MatPlusRand(*Mat + Rand);
  Vectors U_Linearized(*U_m);
  DGSeismogram s_linearized(Obs->GetSpaceOp().GetParaDat());

  LinOp->SolveAndMeasure(Rand, *Obs, s_linearized, &U_Linearized);
  Vectors U_plusMat(*U_m);
  DGSeismogram s(Obs->GetSpaceOp().GetParaDat());
  std::vector<double> quotients(6, 0.0);
  for (int i = 0; i < 6; ++i) {
    double fac = pow(10, -i);
    MatPlusRand = *Mat;
    MatPlusRand += fac * Rand;
    U_plusMat.Clear();
    NonlinearEvaluation(MatPlusRand,U_plusMat,s);
    if (GetParam().Measure) {
      s -= *s_m;
      DGSeismogram tmp(s_linearized);
      tmp *= fac;
      s -= tmp;
      quotients[i] = sqrt(s * s) / fac;
    } else {
      U_plusMat -= *U_m;
      U_plusMat -= fac * U_Linearized;
      double DiffNorm = nc::L2TNorm(U_plusMat, eleCache->Get(), Conf->TConf.dt);
      quotients[i] = DiffNorm / fac;
    }
  }
  MOUT(quotients)
  EXPECT_VECTOR_NEAR(quotients, GetParam().values, 5e-6);
}

INSTANTIATE_TEST_SUITE_P(FWITestLinearized, FWITestLinearized, Values(
    TestParams{"L2","native", 0, false,
               {0.858051,
                0.150326,
                0.0153581,
                0.00153939,
                0.00016032,
                4.63694e-05}},
    TestParams{"L2","native", 1, false,
               {  0.00239439,
                  0.000240443,
                  2.41191e-05,
                  2.88779e-06,
                  1.66784e-06,
                  8.84759e-06}},
    TestParams{"L2","native", 2, false,
               {0.00131595,
                   0.000135594,
                   1.36651e-05,
                   2.06e-06,
                   1.64346e-06,
                   8.84381e-06,
               }},
    TestParams{"L2","native", 0, true,
               {2.47968,
                   0.459557,
                   0.0469712,
                   0.00470215,
                   0.00046857,
                   5.29088e-05}},
    TestParams{"L2","native", 1, true,
               { 0.00724137,
                   0.000727147,
                   7.26582e-05,
                   7.23428e-06,
                   1.19488e-06,
                   1.17118e-05,
               }},
    TestParams{"L2","native", 2, true,
               {  0.00399668,
                   0.000411705,
                   4.11852e-05,
                   4.09754e-06,
                   9.75335e-07,
                   1.16911e-05}},
    TestParams{"L2","taup", 0, false,
               {  0.0475855,
                   0.00638957,
                   0.000666068,
                   0.000107693,
                   8.58458e-05,
                   8.57344e-05}},
    TestParams{"L2","taup", 1, false,
               {   0.650701,
                   0.0962016,
                   0.00974477,
                   0.000976246,
                   0.000103684,
                   3.61869e-05}},
    TestParams{"L2","taup", 2, false,
               {   1.28039,
                   0.407468,
                   0.0512041,
                   0.00525232,
                   0.000529535,
                   8.18302e-05}},
    TestParams{"L2","taup", 0, true,
               {   0.112609,
                   0.0156922,
                   0.00162524,
                   0.000165397,
                   4.96866e-05,
                   4.85602e-05}},
    TestParams{"L2","taup", 1, true,
               {  1.93813,
                   0.294671,
                   0.0298626,
                   0.00298697,
                   0.000297207,
                   3.58037e-05}},
    TestParams{"L2","taup", 2, true,
               {   3.17913,
                   1.07021,
                   0.135763,
                   0.0139414,
                   0.00139907,
                   0.000142399}},
    TestParams{"L2","taup_positive", 0, false,
               {    0.269564,
                   0.0263909,
                   0.00262668,
                   0.000311052,
                   0.000172954,
                   0.000171335}},
    TestParams{"L2","taup_positive", 1, false,
               {   3.6462,
                   2.18105,
                   0.24307,
                   0.0243559,
                   0.00244333,
                   0.000299199}},
    TestParams{"L2","taup_positive", 2, false,
               {    0.0270267,
                   0.00252375,
                   0.00025007,
                   2.5755e-05,
                   6.99414e-06,
                   7.17231e-06}},
    TestParams{"L2","taup_positive", 0, true,
               {   0.613597,
                   0.0657879,
                   0.00655648,
                   0.000652867,
                   0.000111368,
                   9.70708e-05}},
    TestParams{"L2","taup_positive", 1, true,
               {   9.47269,
                   6.62738,
                   0.745344,
                   0.0746819,
                   0.00745923,
                   0.000741845}},
    TestParams{"L2","taup_positive", 2, true,
               {   0.0656569,
                   0.0059467,
                   0.000587198,
                   5.88717e-05,
                   6.94876e-06,
                   6.30177e-06}}
));

int main(int argc, char **argv) {
  return MppTest(
      MppTestBuilder(argc, argv).
          WithoutDefaultConfig().
          WithScreenLogging().
          WithConfPath("../").
          WithGeoPath("../conf/geo/").
          WithPlotPath("data/vtu/").
          WithPPM()
  ).RUN_ALL_MPP_TESTS();
}
