//
// Created by cr on 14.07.21.
//

#include "DGVAcousticMatrixAssembler.hpp"
#include "LagrangeNodalPoints.hpp"

void
DGVAcousticMatrixAssembler::assembleBoundaryFlux(DGRowEntries &A_c, const Mesh &M, const Cell &C,
                                                 const DGAcousticFaceElement &felem,
                                                 bool fwd) const {
  int f = felem.FaceID();
  BFParts bnd(M, C);
  for (int q = 0; q < felem.nQ(); ++q) {
    if (bnd[f] == 1) assembleDirichletBoundaryFlux(A_c, C, felem, q, fwd);
    else if (bnd[f] == 0 || bnd[f] == 2) assembleNeumannBoundaryFlux(A_c, C, felem, q, fwd);
    else if (bnd[f] == 3) assembleRobinBoundaryFlux(A_c, C, felem, q, fwd);
  }
}

void DGVAcousticMatrixAssembler::assembleRobinBoundaryFlux(DGRowEntries &A_c, const Cell &C,
                                                           const DGAcousticFaceElement &felem,
                                                           int q, bool fwd) const {
  double w = felem.QWeight(q);
  VectorField N = felem.QNormal(q);
  Point z = felem.QPoint(q);
  int nV = felem.V_dimension();
  int nP = felem.shape_size();
  double Z_K = sqrt((*Mat).Kappa(C, z) * (*Mat).Rho(C, z));
  double alpha = upwind ? 1.0 / Z_K : 0.0;
  for (int j = 0; j < nV; ++j) {
    double VN_j = felem.Velocity(q, j) * N;
    for (int i = 0; i < nV; ++i) {
      double VN_i = felem.Velocity(q, i) * N;
      A_c(i, j) -= w * Z_K * VN_i * VN_j;
    }
    for (int i = 0; i < felem.shape_size(); ++i) {
      double P_i = felem.Pressure(q, i);
      for (int n = 0; n < felem.p_dimension(); ++n)
        A_c(nV + n * nP + i, j) -= (fwd ? 1.0 : (-1.0)) * w * VN_j * P_i;
    }
  }
  for (int j = 0; j < nP; ++j) {
    double P_j = felem.Pressure(q, j);
    for (int i = 0; i < nV; ++i) {
      double VN_i = felem.Velocity(q, i) * N;
      for (int n = 0; n < felem.p_dimension(); ++n)
        A_c(i, nV + n * nP + j) += w * VN_i * P_j;
    }
    for (int i = 0; i < nP; ++i) {
      double P_i = felem.Pressure(q, i);
      for (int n = 0; n < felem.p_dimension(); ++n)
        for (int k = 0; k < felem.p_dimension(); ++k)
          A_c(nV + n * nP + i, nV + k * nP + j) -=
              w * alpha * P_i * P_j;
    }
  }
}

void DGVAcousticMatrixAssembler::assembleNeumannBoundaryFlux(DGRowEntries &A_c, const Cell &C,
                                                             const DGAcousticFaceElement &felem,
                                                             int q, bool fwd) const {
  int nV = felem.V_dimension();
  int nP = felem.shape_size();
  double w = felem.QWeight(q);
  VectorField N = felem.QNormal(q);
  Point z = felem.QPoint(q);
  double Z_K = sqrt((*Mat).Kappa(C, z) * (*Mat).Rho(C, z));
  for (int j = 0; j < nV; ++j) {
    double VN_j = felem.Velocity(q, j) * N;
    for (int i = 0; i < nV; ++i) {
      double VN_i = felem.Velocity(q, i) * N;
      A_c(i, j) -= w * Z_K * VN_i * VN_j;
    }
    for (int i = 0; i < felem.shape_size(); ++i) {
      double P_i = felem.Pressure(q, i);
      for (int n = 0; n < felem.p_dimension(); ++n) {
        A_c(nV + n * nP + i, j) -= (fwd ? 1.0 : (-1.0)) * w * VN_j * P_i;
      }
    }
  }
}

void DGVAcousticMatrixAssembler::assembleDirichletBoundaryFlux(DGRowEntries &A_c, const Cell &C,
                                                               const DGAcousticFaceElement &felem,
                                                               int q, bool fwd) const {
  double w = felem.QWeight(q);
  VectorField N = felem.QNormal(q);
  Point z = felem.QPoint(q);
  int nV = felem.V_dimension();
  int nP = felem.shape_size();
  double alpha = upwind ? 1.0 / sqrt((*Mat).Kappa(C, z) * (*Mat).Rho(C, z)) : 0.0;
  for (int j = 0; j < nP; ++j) {
    double P_j = felem.Pressure(q, j);
    for (int i = 0; i < nV; ++i) {
      double VN_i = felem.Velocity(q, i) * N;
      for (int n = 0; n < felem.p_dimension(); ++n) {
        A_c(i, nV + n * nP + j) -= (fwd ? 1.0 : (-1.0)) * w * VN_i * P_j;
      }
    }
    for (int i = 0; i < nP; ++i) {
      double P_i = felem.Pressure(q, i);
      for (int n = 0; n < felem.p_dimension(); ++n) {
        for (int k = 0; k < felem.p_dimension(); ++k) {
          A_c(nV + n * nP + i, nV + k * nP + j) -=
              w * alpha * P_i * P_j;
        }
      }
    }
  }
}

void
DGVAcousticMatrixAssembler::assembleFlux(Matrix &A_loc, const DGAcousticElement &elem,
                                         DGRowEntries &A_c, bool fwd) const {
  const Cell &C = elem.GetCell();
  for (int f = 0; f < C.Faces(); ++f) {
    DGAcousticFaceElement felem(A_loc, C, f, (*Mat).GetNumDamping());
    int nV = felem.V_dimension();
    int nP = felem.shape_size();
    int pComponents = (*Mat).GetNumDamping() + 1;
    if (A_loc.OnBoundary(C, f)) {
      if (!dampingFlux) { assembleBoundaryFlux(A_c, A_loc.GetMesh(), C, felem, fwd); }
    } else { // for the rest of the faces
      cell cf = A_loc.find_neighbour_cell(C, f);
      const Cell &CF = *cf;
      int f1 = A_loc.find_neighbour_face_id(C.Face(f), cf);
      DGAcousticFaceElement felem_1(A_loc, CF, f1, (*Mat).GetNumDamping());
      int nV_1 = felem_1.V_dimension();
      int nP_1 = felem_1.shape_size();
      DGRowEntries A_cf(A_loc, C, CF);
      for (int q = 0; q < felem.nQ(); ++q) {
        double w = felem.QWeight(q);
        VectorField N = felem.QNormal(q);
        Point z = felem.QPoint(q);
        double Z_K = sqrt((*Mat).Kappa(C, z) *
                          (*Mat).Rho(C, z));
        double Z_Kf = sqrt((*Mat).Kappa(*cf, z) *
                           (*Mat).Rho(*cf, z));
        double alpha1 = 1.0 / (Z_K + Z_Kf);
        double alpha2 = Z_Kf * Z_K / (Z_K + Z_Kf);
        double alpha3 = (fwd ? 1.0 : -1.0) * Z_Kf / (Z_K + Z_Kf);
        double alpha4 = (fwd ? 1.0 : -1.0) * Z_K / (Z_K + Z_Kf);
        if (!upwind) {
          alpha1 = alpha2 = 0;
          alpha3 = alpha4 = -0.5;
        }
        for (int i = 0; i < nV; ++i) {
          double VN_i = felem.Velocity(q, i) * N;
          for (int j = 0; j < nV; ++j) {
            double VN_j = felem.Velocity(q, j) * N;
            A_c(i, j) -= w * alpha2 * VN_i * VN_j;
          }
          for (int j = 0; j < nV_1; ++j) {
            double V1N_j = felem_1.Velocity(q, j) * N;
            A_cf(i, j) += w * alpha2 * VN_i * V1N_j;
          }
          for (int j = 0; j < nP; ++j) {
            double P_j = felem.Pressure(q, j);
            for (int n = 0; n < pComponents; ++n) {
              A_c(i, nV + n * nP + j) -= w * alpha4 * P_j * VN_i;
            }
          }
          for (int j = 0; j < nP_1; ++j) {
            double P1_j = felem_1.Pressure(q, j);
            for (int n = 0; n < pComponents; ++n) {
              A_cf(i, nV_1 + n * nP_1 + j) += w * alpha4 * P1_j * VN_i;
            }
          }
        }
        for (int i = 0; i < nP; ++i) {
          double P_i = felem.Pressure(q, i);
          for (int j = 0; j < nP; ++j) {
            double P_j = felem.Pressure(q, j);
            for (int n = 0; n < pComponents; ++n) {
              for (int k = 0; k < pComponents; ++k) {
                A_c(nV + n * nP + i, nV + k * nP + j) -=
                    w * alpha1 * P_i * P_j;
              }
            }
          }
          for (int j = 0; j < nP_1; ++j) {
            double P1_j = felem_1.Pressure(q, j);
            for (int n = 0; n < pComponents; ++n) {
              for (int k = 0; k < pComponents; ++k) {
                A_cf(nV + n * nP + i, nV_1 + k * nP + j) +=
                    w * alpha1 * P_i * P1_j;
              }
            }
          }
          for (int j = 0; j < nV; ++j) {
            double VN_j = felem.Velocity(q, j) * N;
            for (int n = 0; n < pComponents; ++n) {
              A_c(nV + n * nP + i, j) -= w * alpha3 * VN_j * P_i;
            }
          }
          for (int j = 0; j < nV_1; ++j) {
            double V1N_j = felem_1.Velocity(q, j) * N;
            for (int n = 0; n < pComponents; ++n) {
              A_cf(nV + n * nP + i, j) += w * alpha3 * V1N_j * P_i;
            }
          }
        }
      }
    }
  }
}

void DGVAcousticMatrixAssembler::FluxMatrix(Matrix &A, bool damping, bool fwd) const {
  A = 0;
  for (const auto &elemIt: *EleCache) {
    const auto &elem = *elemIt.second;
    const auto &r = elem.Row();
    DGRowEntries A_c(A, elem.GetCell(), elem.GetCell());
    assembleDifferentialPartFlux(A, elem, A_c, damping, fwd);
    assembleFlux(A, elem, A_c, fwd);
  }
}

void DGVAcousticMatrixAssembler::MassMatrixOnlyKappa(Matrix &M) const {
  M = 0;
  for (const auto &elemIt: *EleCache) {
    const auto &elem = *elemIt.second;
    const Cell &C = elem.GetCell();
    DGRowEntries M_c(M, C, C);
    int nV = elem.V_dimension();
    int nP = elem.shape_size();
    int p_dim = elem.p_dimension();
    for (int q = 0; q < elem.nQ(); ++q) {
      double w = elem.QWeight(q);
      Point z = elem.QPoint(q);
      for (int i = 0; i < nV; ++i) {
        VectorField V_i = elem.Velocity(q, i);
        for (int j = 0; j < nV; ++j) {
          VectorField V_j = elem.Velocity(q, j);
          M_c(i, j) += w * (V_i * V_j);
        }
      }
      for (int i = 0; i < nP; ++i) {
        double P_i = elem.Pressure(q, i);
        for (int j = 0; j < nP; ++j) {
          double P_j = elem.Pressure(q, j);
          for (int n = 0; n < p_dim; ++n) {
            M_c(i + nV + n * nP, j + nV + n * nP) +=
                w * (*Mat).invKappa_i(C, z, n) * P_i * P_j;
          }
        }
      }
    }
  }
}

void DGVAcousticMatrixAssembler::MassMatrix(Matrix &M) const {
  M = 0;
  for (const auto &elemIt: *EleCache) {
    const auto& elem = *elemIt.second;
    const Cell &C = elem.GetCell();
    DGRowEntries M_c(M, C, C);
    int nV = elem.V_dimension();
    int nP = elem.shape_size();
    int p_dim = elem.p_dimension();
    for (int q = 0; q < elem.nQ(); ++q) {
      double w = elem.QWeight(q);
      Point z = elem.QPoint(q);
      double rho = (*Mat).Rho(C, z);
      for (int i = 0; i < nV; ++i) {
        VectorField V_i = elem.Velocity(q, i);
        for (int j = 0; j < nV; ++j) {
          VectorField V_j = elem.Velocity(q, j);
          M_c(i, j) += w * rho * (V_i * V_j);
        }
      }
      for (int i = 0; i < nP; ++i) {
        double P_i = elem.Pressure(q, i);
        for (int j = 0; j < nP; ++j) {
          double P_j = elem.Pressure(q, j);
          for (int n = 0; n < p_dim; ++n) {
            M_c(i + nV + n * nP, j + nV + n * nP) +=
                w * (*Mat).invKappa_i(C, z, n) * P_i * P_j;
          }
        }
      }
    }
  }
}

void DGVAcousticMatrixAssembler::assembleDifferentialPartFlux(const Matrix &A,
                                                              const DGAcousticElement &elem,
                                                              DGRowEntries &A_c, bool damping,
                                                              bool fwd) const {
  if (damping) assembleDamping(elem, A_c);
  assembleGradP(elem, A_c, fwd);
  assembleDivV(elem, A_c, fwd);
}

void DGVAcousticMatrixAssembler::assembleGradP(const DGAcousticElement &elem, DGRowEntries &A_c,
                                               bool fwd) const {
  int nV = elem.V_dimension();
  int nP = elem.shape_size();
  int p_dim = elem.p_dimension();
  for (int q = 0; q < elem.nQ(); ++q) {
    double w = elem.QWeight(q);
    const Point &z = elem.QPoint(q);
    for (int i = 0; i < nP; ++i) {
      const VectorField &GradP_i = elem.Gradpressure(q, i);
      for (int j = 0; j < nV; ++j) {
        const VectorField &V_j = elem.Velocity(q, j);
        for (int n = 0; n < p_dim; ++n)
          A_c(j, nV + n * nP + i) += (fwd ? 1.0 : (-1.0)) * w * (GradP_i * V_j);
      }
    }
  }
}

void DGVAcousticMatrixAssembler::assembleDivV(const DGAcousticElement &elem, DGRowEntries &A_c,
                                              bool fwd) const {
  int nV = elem.V_dimension();
  int nP = elem.shape_size();
  int p_dim = elem.p_dimension();
  for (int q = 0; q < elem.nQ(); ++q) {
    double w = elem.QWeight(q);
    const Point &z = elem.QPoint(q);
    for (int i = 0; i < nV; ++i) {
      double Div_i = elem.Divergence(q, i);
      for (int j = 0; j < nP; ++j) {
        double P_j = elem.Pressure(q, j);
        for (int n = 0; n < p_dim; ++n)
          A_c(nV + n * nP + j, i) += (fwd ? 1.0 : (-1.0)) * w * Div_i * P_j;
      }
    }
  }
}

void
DGVAcousticMatrixAssembler::assembleDamping(const DGAcousticElement &elem,
                                            DGRowEntries &A_c) const {
  int nV = elem.V_dimension();
  int nP = elem.shape_size();
  int p_dim = elem.p_dimension();
  for (int q = 0; q < elem.nQ(); ++q) {
    double w = elem.QWeight(q);
    const Point &z = elem.QPoint(q);
    for (int i = 0; i < elem.shape_size(); ++i) {
      for (int j = 0; j < nP; ++j) {
        for (int n = 1; n < p_dim; ++n) {
          A_c(nV + n * nP + j, nV + n * nP + i) -=
              w * elem.Pressure(q, i) * elem.Pressure(q, j) *
              (*Mat).invKappa_i(elem.GetCell(), z, n)
              * (*Mat).invTau_i(elem.GetCell(), z, n);
        }
      }
    }
  }
}

void DGVAcousticMatrixAssembler::DampingMatrix(Matrix &D) const {
  D = 0;
  for (const auto &elemIt: *EleCache) {
    const auto &elem = *elemIt.second;
    DGRowEntries D_c(D, elem.GetCell(), elem.GetCell());
    assembleDamping(elem,D_c);
  }
}
